@@ -33,7 +33,7 @@ from googleapiclient.discovery import build
 from googleapiclient.errors import HttpError
 
 from .sqlhelper import _get_config, _get_credentials
-from .misc import file_age, verbose_display, _pycof_folders
+from .misc import file_age, verbose_display, _pycof_folders, EmailSSHTunnel
 
 # Define default Google API scopes
 default_scopes = ['https://mail.google.com/', 'https://www.googleapis.com/auth/calendar.readonly']
@@ -42,7 +42,7 @@ default_scopes = ['https://mail.google.com/', 'https://www.googleapis.com/auth/c
 #######################################################################################################################
 
 # Send an Email
-def send_email(to, subject, body, cc='', credentials={}):
+def send_email(to, subject, body, cc='', credentials={}, connection='auto'):
     """Simplified function to send emails.
     Will look at the credentials at :obj:`/etc/.pycof/config.json`. User can also pass a dictionnary for credentials.
 
@@ -77,10 +77,14 @@ def send_email(to, subject, body, cc='', credentials={}):
         >>> pycof.send_email(to="test@domain.com", body=content, subject="Hello world!")
     """
     config = _get_config(credentials)
+
+    if connection.lower() == 'auto':
+        pattern = re.compile("^([0-9]+.[0-9]+.[0-9]+.[0-9]+)+$")
+        connection = 'direct' if pattern.match(config.get('EMAIL_SMTP')) else 'ssh'
     msg = MIMEMultipart()
-    msg['From'] = config.get('EMAIL_SENDER')
+    msg['From'] = config.get('EMAIL_USER')
     msg['To'] = to
-    msg['Cc'] = '' if cc == '' else cc
+    # msg['Cc'] = '' if cc == '' else cc
     msg['Subject'] = subject
 
     mail_type = 'html' if '</' in body else 'plain'
@@ -88,19 +92,11 @@ def send_email(to, subject, body, cc='', credentials={}):
 
     text = msg.as_string()
 
-    # Server login
-    try:
-        port = str(config.get('EMAIL_PORT'))
-    except Exception:
-        port = '587'  # Default Google port number
-    connection = config.get('EMAIL_SMTP') + ':' + port
-    server = smtplib.SMTP(connection)
-    server.starttls()
-    server.login(user=config.get('EMAIL_USER'), password=config.get('EMAIL_PASSWORD'))
-
-    # Send email
-    server.sendmail(config.get('EMAIL_USER'), [to, '', cc], text)
-    server.quit()
+    with EmailSSHTunnel(config=config, connection=connection) as tunnel:
+        conn = tunnel.connector()
+        conn.sendmail(config.get('EMAIL_USER'), [to, '', cc], text)
+        conn.quit()
+        # conn.sendmail('noreply@florianfelice.com','florian@florianfelice.com',msg.as_string())
 
 # Send an email from Gmail
 class google_email:
@@ -196,7 +192,7 @@ class google_email:
         if return_status:
             return send_message
 
-    def getEmails(self, nb_email=1, _to='me'):
+    def getEmails(self, nb_email=1, include_spam=False, _to='me', *args):
         """Get latest emails from your Gmail address.
 
         :param nb_email: Number of emails to retreive, defaults to 1.
@@ -205,7 +201,7 @@ class google_email:
         :type _to: :obj:`str`, optional
 
         :example:
-            >>> pycof.GetEmails(2)
+            >>> pycof.google_emails().GetEmails(2)
             ... +----------------------------+-----------------+----------------+---------------+
             ... |                       Date |            From |        Subject |            To |
             ... +----------------------------+-----------------+----------------+---------------+
@@ -222,7 +218,7 @@ class google_email:
         service = build('gmail', 'v1', credentials=creds)
 
         # request a list of all the messages
-        result = service.users().messages().list(userId=_to).execute()
+        result = service.users().messages().list(userId=_to, maxResults=nb_email, includeSpamTrash=include_spam, *args).execute()
 
         # We can also pass maxResults to get any number of emails. Like this:
         # result = service.users().messages().list(maxResults=200, userId='me').execute()
@@ -243,13 +239,17 @@ class google_email:
                 payload = txt['payload']
                 headers = payload['headers']
 
+                _dest = _to
+                _subj = ''
+                _from = ''
+
                 # Look for Subject and Sender Email in the headers
                 for d in headers:
-                    if d['name'] == 'subject':
+                    if d['name'].lower() == 'subject':
                         _subj = d['value']
                     if d['name'] == 'From':
                         _from = d['value']
-                    if d['name'] == 'to':
+                    if d['name'].lower() == 'to':
                         _dest = d['value']
                     if d['name'] == 'Date':
                         ddt = dateparser.parse(d['value'].strip())
@@ -263,32 +263,35 @@ class google_email:
                 attch = {}
                 _body = ''
                 for part in payload.get('parts'):
-                    if part['filename']:
-                        att_id = part['body']['attachmentId']
-                        att = service.users().messages().attachments().get(userId=_to, messageId=msg['id'], id=att_id).execute()
-                        data = att['data']
-                        file_data = base64.urlsafe_b64decode(data.encode('UTF-8'))
-                        filePath = os.path.join(_pycof_folders('data'), part['filename'])
-                        attch.update({f'Attachment {i}': filePath})
-                        # Download the attachment
-                        with open(filePath, 'wb') as f:
-                            f.write(file_data)
-                        i += 1
-                    else:
-                        data = part['body']['data']
-                        data = data.replace("-", "+").replace("_", "/")
-                        decoded_data = base64.b64decode(data)
+                    try:
+                        if part['filename']:
+                            att_id = part['body']['attachmentId']
+                            att = service.users().messages().attachments().get(userId=_to, messageId=msg['id'], id=att_id).execute()
+                            data = att['data']
+                            file_data = base64.urlsafe_b64decode(data.encode('UTF-8'))
+                            filePath = os.path.join(_pycof_folders('data'), part['filename'])
+                            attch.update({f'Attachment {i}': filePath})
+                            # Download the attachment
+                            with open(filePath, 'wb') as f:
+                                f.write(file_data)
+                            i += 1
+                        else:
+                            data = part['body']['data']
+                            data = data.replace("-", "+").replace("_", "/")
+                            decoded_data = base64.b64decode(data)
 
-                        # Now, the data obtained is in lxml. So, we will parse
-                        # it with BeautifulSoup library
-                        soup = BeautifulSoup(decoded_data, "lxml")
-                        _body = soup.body()
+                            # Now, the data obtained is in lxml. So, we will parse
+                            # it with BeautifulSoup library
+                            soup = BeautifulSoup(decoded_data, "lxml")
+                            _body = soup.body()
+                    except Exception:
+                        pass
                 for_df = {'From': _from, 'To': _dest, 'Date': _date, 'Subject': _subj}  # , 'Body': _body
                 for_df.update(attch)
                 df += [pd.DataFrame(for_df, index=[0])]
             except Exception:
                 pass
-        return pd.concat(df).reset_index(drop=True)
+        return pd.concat(df).sort_values(by='Date', ascending=False).reset_index(drop=True)
 
 
 #######################################################################################################################
@@ -685,20 +688,30 @@ def GetEmails(nb_email=1, email_address='', port=993, credentials={}):
         for num in range(latest_email_id, latest_email_id - nb_email, -1):
             typ, data = mail.fetch(str(num), '(RFC822)')
             raw_email = data[0][1]  # converts byte literal to string removing b''
-            raw_email_string = raw_email.decode('utf-8')
+            try:
+                raw_email_string = raw_email.decode('utf-8')
+            except:
+                raw_email_string = ''
             email_message = email.message_from_string(raw_email_string)  # downloading attachments
             # Get email content
-            msg = email.message_from_string(str(raw_email, 'utf-8'))
-            _subj = msg['subject']
-            _from = msg['From']
-            _to = msg['To']
-            ddt = dateparser.parse(msg['Date'].strip())
+            try:
+                msg = email.message_from_string(str(raw_email, 'utf-8'))
+                _subj = msg['subject']
+                _from = msg['From']
+                _to = msg['To']
+                ddt = dateparser.parse(msg['Date'].strip())
+            except:
+                msg = ''
+                ddt = np.nan
+
             if ddt is None:
                 ddt = dateparser.parse(msg['Date'].replace('00 (PST)', ' PST').split(',')[1])
+
             try:
                 _date = ddt.replace(tzinfo=datetime.timezone.utc).astimezone(tz=tz.tzlocal())
             except Exception:
                 _date = np.nan
+
             for_df = {'From': _from, 'Subject': _subj, 'To': _to, 'Date': _date}
 
             # Get email attachments
