@@ -12,20 +12,29 @@ from io import StringIO, BytesIO
 
 from tqdm import tqdm
 import datetime
+import sshtunnel
+
+import smtplib
+from email.mime.multipart import MIMEMultipart
+from email.mime.text import MIMEText
+import logging
+import colorlog
+
 
 
 ########################################################################################################################
 # Get config file
 
 def _pycof_folders(output=None, verbose=False):
+    pycof_fold = os.environ.get('PYCOF_PATH', os.sep)
     # Define the root folder depending on the OS
     if sys.platform in ['win32', 'win64', 'cygwin', 'msys']:
         temp_path = os.environ['TEMP'] + os.sep
         home = os.path.expanduser('~')
         creds_fold = os.path.join(home, '.pycof')  + os.sep
     else:
-        temp_path = os.path.join(os.sep, 'tmp') + os.sep
-        creds_fold = os.path.join(os.sep, 'etc', '.pycof') + os.sep
+        temp_path = os.path.join(pycof_fold, 'tmp') + os.sep
+        creds_fold = os.path.join(pycof_fold, 'etc', '.pycof') + os.sep
         home = os.path.expanduser('~')
 
     # Initialize the creation count variable
@@ -59,6 +68,26 @@ def _pycof_folders(output=None, verbose=False):
         except PermissionError as err:
             raise PermissionError(f"""Could not create the PYCOF temp data folder, permission denied: {folds_d}.
                     Please create folder with: sudo mkdir {folds_d} or run your script with super-user.""")
+    
+    # S3 temp folder
+    folds_s3 = os.path.join(temp_path, 'pycof', 'cache', 's3') + os.sep
+    if not os.path.exists(folds_s3):
+        try:
+            os.makedirs(folds_s3)
+            _created += 1
+        except PermissionError as err:
+            raise PermissionError(f"""Could not create the PYCOF temp s3 folder, permission denied: {folds_s3}.
+                    Please create folder with: sudo mkdir {folds_s3} or run your script with super-user.""")
+    
+    # Models temp folder
+    folds_models = os.path.join(temp_path, 'pycof', 'cache', 'models') + os.sep
+    if not os.path.exists(folds_models):
+        try:
+            os.makedirs(folds_models)
+            _created += 1
+        except PermissionError as err:
+            raise PermissionError(f"""Could not create the PYCOF temp models folder, permission denied: {folds_models}.
+                    Please create folder with: sudo mkdir {folds_models} or run your script with super-user.""")
 
     # Return path if asked by user
     if output in ['tmp', 'temp']:
@@ -71,6 +100,10 @@ def _pycof_folders(output=None, verbose=False):
         return folds_d
     elif output == 'home':
         return home
+    elif output == "s3":
+        return folds_s3
+    elif output == 'models':
+        return folds_models
     elif verbose:
         print(f'PYCOF folder created: {_created}')
 
@@ -181,8 +214,12 @@ def write(file, path, perm='a', verbose=False, end_row='\n', credentials={}, pro
             data_path = os.path.join(_pycof_folders('temp'), 'pycof', 'cache', 'data') + os.sep
             path = data_path + file_name
 
+        
         with open(path, perm) as f:
-            f.write(file + end_row)
+            if path.endswith('.json') or path.endswith('.jsonc'):
+                json.dump(file, f)
+            else:
+                f.write(file + end_row)
 
         if useIAM:
             s3.upload_file(path, bucket, folder_path)
@@ -257,3 +294,128 @@ def verbose_display(element, verbose=True, sep=' ', end='\n', return_list=False)
         disp = 0  # we don't display anything
     else:
         return(element)
+
+
+# #######################################################################################################################
+# Fake SSH tunnel for direct connections
+
+class _fake_tunnel:
+    def __init__():
+        pass
+
+    def close():
+        pass
+
+########################################################################################################################
+# SSHTunnel for email
+class EmailSSHTunnel:
+    def __init__(self, config, connection='direct', engine='default'):
+        self.connection = connection.lower()
+        self.config = config
+        self.engine = engine
+
+    def __enter__(self):
+        if self.connection == 'ssh':
+            try:
+                ssh_port = 22 if self.config.get('SSH_PORT') is None else int(self.config.get('SSH_PORT'))
+                remote_addr = 'localhost' if self.config.get('EMAIL_REMOTE_HOST') is None else self.config.get('EMAIL_REMOTE_HOST')
+                remote_port = 1025 if self.config.get('EMAIL_REMOTE_PORT') is None else int(self.config.get('EMAIL_REMOTE_PORT'))
+                hostname = '127.0.0.1' if self.config.get('EMAIL_LOCAL_HOST') is None else self.config.get('EMAIL_LOCAL_HOST')
+
+                if (self.config.get('SSH_PASSWORD') is None) & (self.config.get('SSH_KEY') is None):
+                    # Try to get the default SSH location if neither a password nor a path is provided
+                    ssh_path = os.path.join(_pycof_folders('home'), '.ssh', 'id_rsa')
+                else:
+                    ssh_path = self.config.get('SSH_KEY')
+
+                self.tunnel = sshtunnel.SSHTunnelForwarder((self.config.get('EMAIL_SMTP'), ssh_port),
+                                                           ssh_username=self.config.get('SSH_USER'),
+                                                           ssh_password=self.config.get('SSH_PASSWORD'),
+                                                           ssh_pkey=ssh_path,
+                                                           remote_bind_address=(remote_addr, remote_port))
+                self.tunnel.daemon_forward_servers = True
+                self.tunnel.connector = self._define_connector
+            except Exception:
+                raise ConnectionError('Failed to establish SSH connection with host')
+        else:
+            self.tunnel = _fake_tunnel
+            self.tunnel.connector = self._define_connector
+
+        return self.tunnel
+
+    def __exit__(self, exc_type, exc_val, exc_tb):
+        self.tunnel.close()
+
+    def _define_connector(self):
+        """Define the SQL connector for executing SQL.
+
+        :param config: Credentials file containing authentiation information, defaults to {}.
+        :type config: :obj:`dict`, optional
+        :param engine: SQL engine to use ('Redshift', 'SQLite' or 'MySQL'), defaults to 'default'
+        :type engine: str, optional
+        :param connection: Connextion type. Cqn either be 'direct' or 'SSH', defaults to 'direct'
+        :type connection: str, optional
+
+        :return: Connector, cursor and tunnel
+        """
+        user = self.config.get('EMAIL_USER')
+        password = self.config.get('EMAIL_PASSWORD')
+        hostname = self.config.get('EMAIL_SMTP')
+        port = self.config.get('EMAIL_PORT')
+
+        if self.connection.lower() == 'ssh':
+            hostname = '127.0.0.1' if self.config.get('EMAIL_REMOTE_HOST') is None else self.config.get('EMAIL_REMOTE_HOST')
+            self.tunnel.start()
+            port = self.tunnel.local_bind_port
+
+        try:
+            connector = smtplib.SMTP(hostname, port)
+            connector.login(user, password)
+        except Exception:
+            raise ConnectionError('Failed to connect to the email server')
+
+        return connector
+
+
+########################################################################################################################
+# Logging setup
+def setup_logging(file=__name__, logging_level=logging.INFO):
+    """
+    Set up logging configuration.
+    """
+    # Create a logger
+    logger = logging.getLogger(file)
+    logger.setLevel(logging_level)
+
+    # Create handlers
+    console_handler = logging.StreamHandler()
+    console_handler.setLevel(logging_level)
+
+    # Create formatters and add them to the handlers
+    formatter = colorlog.ColoredFormatter(
+        "%(log_color)s%(asctime)s [%(levelname)s] %(name)s %(reset)s %(message_log_color)s - %(message)s",
+        datefmt='%Y-%m-%d %H:%M:%S',
+        reset=True,
+        log_colors={
+            'DEBUG': 'light_black',
+            'INFO': 'blue',
+            'WARNING': 'yellow',
+            'ERROR': 'red',
+            'CRITICAL': 'bold_red',
+        },
+        secondary_log_colors={
+            'message': {
+                'DEBUG': 'light_black',
+                'INFO': 'white',
+                'WARNING': 'light_yellow',
+                'ERROR': 'red',
+                'CRITICAL': 'red',
+            }
+        }
+    )
+    console_handler.setFormatter(formatter)
+
+    # Add the handlers to the logger
+    logger.addHandler(console_handler)
+
+    return logger
