@@ -2,7 +2,6 @@ import boto3
 import botocore
 
 import sshtunnel
-import pymysql
 import psycopg2
 import sqlite3
 
@@ -15,20 +14,20 @@ from tqdm import tqdm
 import os
 import sys
 import getpass
-import json
+import sqlalchemy as sa
 import datetime
 import hashlib
 import warnings
 import csv
 from types import SimpleNamespace
 
-from .misc import verbose_display, file_age, write, _get_config, _pycof_folders
-from .data import f_read
+from .misc import verbose_display, file_age, write, _get_config, _pycof_folders, _fake_tunnel
+from .data import read
 
 # #######################################################################################################################
 # Cache data from SQL
 
-def _cache(sql, tunnel, query_type="SELECT", cache_time='24h', cache_file_name=None, verbose=False):
+def _cache(sql, tunnel, query_type="SELECT", cache_time='24h', cache_file_name=None, cache_folder=None, verbose=False):
     # Parse cache_time value
     if type(cache_time) in [float, int]:
         c_time = cache_time
@@ -46,7 +45,7 @@ def _cache(sql, tunnel, query_type="SELECT", cache_time='24h', cache_file_name=N
 
     # Set the query and data paths
     query_path = _pycof_folders('queries')
-    data_path = _pycof_folders('data')
+    data_path = cache_folder if cache_folder else _pycof_folders('data')
 
     # Chec if the cached data already exists
     if (query_type.upper() == "SELECT") & (file_name in os.listdir(data_path)):
@@ -55,38 +54,40 @@ def _cache(sql, tunnel, query_type="SELECT", cache_time='24h', cache_file_name=N
         if (query_type.upper() == "SELECT") & (age < c_time):
             # If file is younger than c_time, we read the cached data
             verbose_display('Reading cached data', verbose)
-            read = f_read(os.path.join(data_path, file_name))
+            sql_out = read(os.path.join(data_path, file_name))
         else:
             # Else we execute the SQL query and save the ouput + the query
             verbose_display('Execute SQL query and cache the data - updating cache', verbose)
             conn = tunnel.connector()
-            read = pd.read_sql(sql, conn)
+            warnings.filterwarnings("ignore", category=UserWarning, message='.*pandas only supports SQLAlchemy connectable.*')
+            sql_out = pd.read_sql(sql, conn)
             conn.close()
             write(sql, query_path + file_name, perm='w', verbose=verbose)
-            write(read, os.path.join(data_path, file_name), index=False)
+            write(sql_out, os.path.join(data_path, file_name), index=False)
     else:
         # If the file does not even exist, we execute SQL, save the query and its output
         verbose_display('Execute SQL query and cache the data', verbose)
         conn = tunnel.connector()
-        read = pd.read_sql(sql, conn)
+        warnings.filterwarnings("ignore", category=UserWarning, message='.*pandas only supports SQLAlchemy connectable.*')
+        sql_out = pd.read_sql(sql, conn)
         conn.close()
         write(sql, os.path.join(query_path, file_name), perm='w', verbose=verbose)
-        write(read, os.path.join(data_path, file_name), index=False)
+        write(sql_out, os.path.join(data_path, file_name), index=False)
 
     def age(fmt='seconds'):
         return file_age(file_path=os.path.join(data_path, file_name), format=fmt)
 
-    read.meta = SimpleNamespace()
-    read.meta.cache = SimpleNamespace()
-    read.meta.cache.creation_date = datetime.datetime.now() - datetime.timedelta(seconds=age())
-    read.meta.cache.cache_path = os.path.join(data_path, file_name)
-    read.meta.cache.query_path = os.path.join(query_path, file_name)
-    read.meta.cache._age_format = age_fmt
-    read.meta.cache._age_value = c_time
-    read.meta.cache._cache_time = cache_time
-    read.meta.cache.age = age
+    sql_out.meta = SimpleNamespace()
+    sql_out.meta.cache = SimpleNamespace()
+    sql_out.meta.cache.creation_date = datetime.datetime.now() - datetime.timedelta(seconds=age())
+    sql_out.meta.cache.cache_path = os.path.join(data_path, file_name)
+    sql_out.meta.cache.query_path = os.path.join(query_path, file_name)
+    sql_out.meta.cache._age_format = age_fmt
+    sql_out.meta.cache._age_value = c_time
+    sql_out.meta.cache._cache_time = cache_time
+    sql_out.meta.cache.age = age
 
-    return read
+    return sql_out
 
 
 # #######################################################################################################################
@@ -144,16 +145,6 @@ def _get_credentials(config, profile_name=None, connection='direct'):
     return config
 
 
-# #######################################################################################################################
-# Fake SSH tunnel for direct connections
-
-class _fake_tunnel:
-    def __init__():
-        pass
-
-    def close():
-        pass
-
 # #######################################################################################################################
 # Get SSH tunnel
 
@@ -239,11 +230,15 @@ class SSHTunnel:
         else:
             try:
                 # Add new encoder of numpy.float64
-                pymysql.converters.encoders[np.float64] = pymysql.converters.escape_float
-                pymysql.converters.conversions = pymysql.converters.encoders.copy()
-                pymysql.converters.conversions.update(pymysql.converters.decoders)
-                # Create connection
-                connector = pymysql.connect(host=hostname, port=int(port), user=user, password=password)
+                # pymysql.converters.encoders[np.float64] = pymysql.converters.escape_float
+                # pymysql.converters.conversions = pymysql.converters.encoders.copy()
+                # pymysql.converters.conversions.update(pymysql.converters.decoders)
+                # # Create connection
+                # connector = pymysql.connect(host=hostname, port=int(port), user=user, password=password)
+
+                params = "{user}:{password}@{hostname}:{port}/".format(user=user, password=password, hostname=hostname, port=port)
+
+                connector = sa.create_engine("mysql+pymysql://{}".format(params)).raw_connection()
             except Exception:
                 raise ConnectionError('Failed to connect to the MySQL database')
 
@@ -270,7 +265,7 @@ def _insert_data(data, table, connector, autofill_nan=False, verbose=False):
     warnings.filterwarnings('ignore')  # Removing filter warning when changing data type
     dt_cols = []
     for col in data.columns:
-        if data[col].dtype == 'object':
+        if type(data.sample(1).reset_index()[col][0]) in [datetime.date, pd._libs.tslibs.timestamps.Timestamp]:
             try:
                 data.loc[:, col] = pd.to_datetime(data[col]).apply(str)
                 dt_cols += [col]
