@@ -5,9 +5,9 @@ with open("README.md", "r") as fh:
 
 setuptools.setup(
     name="pycof",
-    version="1.4.2",
+    version="1.5.13",
     author="Florian Felice",
-    author_email="florian.felice@outlook.com",
+    author_email="admin@florianfelice.com",
     description="A package for commonly used functions",
     long_description=long_description,
     long_description_content_type="text/markdown",
@@ -17,7 +17,7 @@ setuptools.setup(
           "pandas>=0.24.1",
           "numpy>=1.16.3",
           "psycopg2-binary>=2.7.4",
-          "pymysql>=0.9.3",
+          "sqlalchemy>=2.0.42",
           "tqdm>=4.35.0",
           "boto3>=1.16.19",
           "xlrd>=1.2.0",
@@ -28,7 +28,10 @@ setuptools.setup(
           "google-auth-httplib2>=0.0.4",
           "google-auth-oauthlib>=0.4.2",
           "google-auth>=1.24.0",
-          "httplib2>=0.18.1"
+          "httplib2>=0.18.1",
+          "pyarrow>=11.0.0",
+          "bs4>=0.0.1",
+          "colorlog>=6.9.0"
       ],
     classifiers=[
         "Programming Language :: Python :: 3",
